/*
 * GANTS - GNOME Applet for the ANTS system
 *
 * (C) 2000, 2001 by Jakob Oestergaard  (joe@sysorb.com / jakob@unthought.net)
 *
 * $Id: gants.c,v 1.3 2003/04/22 10:28:37 ksm Exp $
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later
 *  version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 *
 */

#include "../src/config.h"
#include <gnome.h>
#include <applet-widget.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <time.h>

#include <errno.h>
#include <string.h>

#include "../src/antslib.h"

#define TEXT_MARGIN 6
#define BAR_MARGIN 2
#define CELL_SIZE 6

typedef struct _metrics_entry_t {
  unsigned peer_id;
  unsigned max, preferred;
  unsigned running, reserved;
  struct _metrics_entry_t * next;
} metrics_entry_t;

static metrics_entry_t * metrics = 0;
static unsigned n_metrics = 0;

static const char * jobtypes[] = {
  "g++", "gcc", "ld", "dep", 0 };

static const char * active_jobtype = 0;
static PanelOrientType orient;
static gint panel_size = 40;

static int usock = -1;
static gint usock_tag;
static gboolean is_realized = FALSE;

static GtkWidget * applet = 0;
static GtkWidget * disp = 0;
static GdkPixmap * pixmap = 0;
static GdkGC * gc = 0;
static GdkColor bg_color, low_color, pre_color, high_color, f_color;
static GdkFont *my_font = 0;

static void load_my_font(void);
static void applet_change_orient(AppletWidget *, PanelOrientType, gpointer);
static void applet_change_pixel_size(AppletWidget *, int, gpointer);
static gint setup_local_socket(gpointer);
static void kill_local_socket(void);
static void handle_local_read(gpointer, gint, GdkInputCondition);
static void setup_colors(void);
static void create_gc(void);

#if 0
static void gants_config(AppletWidget*,gpointer);
#endif

static void show_gants_about(AppletWidget*,gpointer);
static void change_jobtype(AppletWidget* aw, gpointer ap);
static GtkWidget * gants_widget(void);
static gint gants_expose(GtkWidget *widget, GdkEventExpose *event);
static gint gants_configure(GtkWidget *widget, GdkEventConfigure *event);
static void gants_resize(void);
static void gants_draw(void);
static void gants_draw_v(void);
static void gants_draw_h(void);
static int gants_get_best_size_v(void);
static int gants_get_best_size_h(void);


int main(int argc, char **argv)
{
  GtkWidget *gwidget;

  /* Initialize the i18n stuff */
  bindtextdomain("gANTS", GNOMELOCALEDIR);
  textdomain("gANTS");

  /* intialize, this will basically set up the applet, corba and call gnome_init */
  applet_widget_init("gants_applet", 0, argc, argv, 0, 0, 0);

  /* create a new applet_widget */
  applet = applet_widget_new("gants_applet");

  /* in the rare case that the communication with the panel failed, error out */
  if(!applet)
    g_error("Can't create applet!\n");

  /* create the widget we are going to put on the applet */
  gwidget = gants_widget();

  gtk_widget_show(gwidget);

  /* add the widget to the applet-widget, and thereby actually putting it "onto" the panel */
  applet_widget_add (APPLET_WIDGET(applet), gwidget);

  /* connect signals */
  gtk_signal_connect(GTK_OBJECT(applet),"change_orient",
		     GTK_SIGNAL_FUNC(applet_change_orient),
		     0);
  gtk_signal_connect(GTK_OBJECT(applet),"change_pixel_size",
		     GTK_SIGNAL_FUNC(applet_change_pixel_size),
		     0);

  /* show & stuff */
  gtk_widget_show_all(applet);
  is_realized = TRUE;

  /* add an item to the applet menu */
  applet_widget_register_stock_callback(APPLET_WIDGET(applet),
					"about",
					GNOME_STOCK_MENU_ABOUT,
					_("About..."),
					show_gants_about, 0);

#if 0
  applet_widget_register_callback(APPLET_WIDGET(applet),
				  "config",
				  _("Configuration..."),
				  &gants_config, 0);
#endif

  /* have the socket set up */
  gtk_timeout_add(1000, setup_local_socket, 0);

  /* setup colors */
  setup_colors();

  /* special corba main loop */
  applet_widget_gtk_main();

  return 0;
}

static void load_my_font(void)
{
  if(!my_font)
    my_font = gdk_font_load ("fixed");
}

static void applet_change_orient(AppletWidget *w, PanelOrientType o, gpointer data)
{
  orient = applet_widget_get_panel_orient (w);
  gants_resize();
}

static void applet_change_pixel_size(AppletWidget *w, int size, gpointer data)
{
  panel_size = applet_widget_get_panel_pixel_size (w);
  gants_resize();
}


static gint setup_local_socket(gpointer p)
{
  /*
   * Make sure we're connected to local antsd
   */
  if(usock == -1) {
    int rc;

    usock = ants_local_connect();
    if(usock < 0) {
      usock = -1;
      /* bad ! */
      g_print(_("Yuck! connect() failed\n"));
      return TRUE;
    }

    rc = ants_statclient(usock);
    if(rc != 0) {
      g_print(_("Couldn't tell ANTS that we're a statclient"));
      close(usock);
      usock = -1;
      return TRUE;
    }
  }

  /*
   * Now remember that we want to handle data sent to us
   */
  usock_tag = gdk_input_add(usock, GDK_INPUT_READ, handle_local_read, 0);

  g_print(_("Added socket\n"));
  gants_draw();

  return FALSE;
}


/*
 * A routine that will correctly clean up after the socket stuff
 */
static void kill_local_socket(void)
{
  if(usock == -1) return;

  gdk_input_remove(usock_tag);
  close(usock);

  gtk_timeout_add(1000, setup_local_socket, 0);

  g_print(_("Kill sock\n"));

  usock = -1;

  gants_draw();
}


/*
 * Called whenever we can read 
 */
static void handle_local_read(gpointer data, gint source, GdkInputCondition cond)
{
  static char * buf = 0;
  int rc;

  unsigned peer_id;
  char jobtype[TYPE_LEN];
  unsigned max, preferred;
  unsigned running, reserved;

  metrics_entry_t * mi;

  /*
   * Make sure we're supposed to read...
   */
  if(cond != GDK_INPUT_READ) {
    kill_local_socket();
    return;
  }

  /*
   * Check buffer alloc.
   */
  if(!buf) {
    buf = (char*)malloc(LOG_LEN);
    if(!buf) return;
  }

  /*
   * Read and parse
   */
  rc = recv(usock, buf, LOG_LEN, 0);
  if(rc <= 0) {
    g_print(_("Yuck! recv() failed\n"));
    kill_local_socket();
    return;
  }

  rc = sscanf(buf, "%u %15s %u %u %u %u",
	      &peer_id, jobtype, 
	      &max, &preferred,
	      &running, &reserved);
  if(rc != 6) {
    g_print(_("Yuck! sscanf() failed\n"));
    return;
  }

  /*
   * If it's not the job type we're interested
   * in, ignore...
   */
  if(!active_jobtype) {
    return;
  }
  if( strcmp(jobtype, active_jobtype) ) {
    return;
  }

  /*
   * Now find metric in list and update values,
   * or insert it if it's new.
   */
  mi = metrics;
  while(mi && mi->peer_id != peer_id) {
    mi = mi->next;
  }

  /* If not found - insert */
  if(!mi) {
    /* allocate new entry */
    mi = (metrics_entry_t*)malloc(sizeof(metrics_entry_t));
    if(!mi) {
      g_print("Yuck! malloc() failed!\n");
      return;
    } 
    /* Be sure to resize - we're extending now... */
    n_metrics++;   
    gants_resize();
    /* chain into list */
    mi->next = metrics;
    metrics = mi;
    /* set std. values */
    mi->peer_id = peer_id;
  }

  /* Update record */
  mi->max = max;
  mi->preferred = preferred;
  mi->running = running;
  mi->reserved = reserved;

  /*
   * Redraw
   */
  gants_draw();

}


static void setup_colors(void)
{
  GdkColormap *colormap;

  colormap = gtk_widget_get_colormap(disp);
                
  gdk_color_parse("black", &bg_color);
  gdk_color_alloc(colormap, &bg_color);

  gdk_color_parse("green", &low_color);
  gdk_color_alloc(colormap, &low_color);
        
  gdk_color_parse("blue", &pre_color);
  gdk_color_alloc(colormap, &pre_color);
	
  gdk_color_parse("red", &high_color);
  gdk_color_alloc(colormap, &high_color);

  gdk_color_parse("gray", &f_color);
  gdk_color_alloc(colormap, &f_color);
}
	        
static void create_gc(void)
{
  gc = gdk_gc_new(disp->window);
  gdk_gc_copy(gc, disp->style->white_gc);
}

static void change_jobtype(AppletWidget* aw, gpointer ap)
{
  static int ajt = 0;

  if(!(active_jobtype = jobtypes[++ajt]))
    active_jobtype = jobtypes[ajt = 0];

  while(metrics) {
    metrics_entry_t * next = metrics->next;
    free(metrics);
    metrics = next;
  }
  metrics = 0;
  n_metrics = 0;

  gants_draw();
}

#if 0
/*
 * Configuration dialog
 */
static void gants_config(AppletWidget* aw, gpointer ap)
{
  GtkWidget * cwind;

  cwind = gtk_window_new(GTK_WINDOW_DIALOG);

  gtk_widget_show(cwind);
}
#endif

/*
 * The mandatory about box  ;)
 */
static void show_gants_about(AppletWidget* aw, gpointer ap)
{
  static GtkWidget *about = 0;
  const gchar *authors[] = {
    "Jakob Oestergaard <jakob@unthought.net>", 0
  };

  if(about) {
    gdk_window_show(about->window);
    gdk_window_raise(about->window);
    return;
  }

  about = gnome_about_new (_("ANTS Activity Applet"), ANTS_VERSION,
			   "(C) 2000,2001",
			   authors,
			   _("Released under the GNU general public license.\n"
			     "Monitors the current ANTS job activity on a cluster"),
			   0);
  gtk_signal_connect(GTK_OBJECT (about), "destroy",
		     GTK_SIGNAL_FUNC (gtk_widget_destroyed), &about);

  gtk_widget_show(about);
}


static GtkWidget * gants_widget(void)
{
  GtkWidget * frame;
  GtkWidget * box;

  box = gtk_vbox_new(FALSE, FALSE);
  gtk_widget_show(box);

  frame = gtk_frame_new(0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);

  disp = gtk_drawing_area_new();
  gtk_signal_connect(GTK_OBJECT(disp), "expose_event",
		     GTK_SIGNAL_FUNC(gants_expose), 0);
  gtk_signal_connect(GTK_OBJECT(disp), "configure_event",
		     GTK_SIGNAL_FUNC(gants_configure), 0);
  gtk_widget_set_events(disp, GDK_EXPOSURE_MASK);

  gtk_widget_set_events(disp, GDK_BUTTON_PRESS_MASK);
  gtk_signal_connect(GTK_OBJECT(disp), "button_press_event",
		     GTK_SIGNAL_FUNC(change_jobtype), 0);

  gtk_box_pack_start_defaults(GTK_BOX(box), disp);

  gtk_container_add(GTK_CONTAINER (frame), box);
	
  orient = applet_widget_get_panel_orient(APPLET_WIDGET(applet));
  panel_size = applet_widget_get_panel_pixel_size(APPLET_WIDGET(applet));

  gants_resize();

  gtk_widget_show_all(frame);

  return frame;
}

static gint gants_expose(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
		  pixmap,
  		  event->area.x, event->area.y,
  		  event->area.x, event->area.y,
		  event->area.width, event->area.height);
  return FALSE;
}

static gint gants_configure(GtkWidget *widget, GdkEventConfigure *event)
{
  if(pixmap) {
    gdk_pixmap_unref(pixmap);
  }

  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height,
			  gtk_widget_get_visual(disp)->depth);
  gdk_pixmap_ref(pixmap);

  gants_draw();

  return TRUE;
} 

static void gants_resize(void)
{
  if ((orient == ORIENT_LEFT) || (orient == ORIENT_RIGHT)) {
    gtk_widget_set_usize(applet, panel_size, gants_get_best_size_v());
    gtk_widget_set_usize(disp, panel_size, gants_get_best_size_v());
  } else {
    gtk_widget_set_usize(applet, gants_get_best_size_h(), panel_size);
    gtk_widget_set_usize(disp, gants_get_best_size_h(), panel_size);
  }
}

static void gants_draw(void)
{
  if(!gc) {
    /* create gc */
    create_gc();
  }

  if(usock == -1) {
    /*
     * Tell the user antsd is down
     */
    gchar tbuf[256];
    gint text_height;

    load_my_font();

    gdk_gc_set_foreground(gc, &bg_color);
    gdk_draw_rectangle(pixmap, gc, TRUE, 0,0,
		       disp->allocation.width,
		       disp->allocation.height);

    gdk_gc_set_foreground(gc, &f_color);
  
    strcpy(tbuf, "ANTSD");
    text_height = gdk_string_height(my_font, tbuf);
    gdk_draw_string(pixmap, my_font, gc,
		    TEXT_MARGIN/2, 
		    TEXT_MARGIN/2 + text_height,
		    tbuf);
    strcpy(tbuf, "Down!");
    gdk_draw_string(pixmap, my_font, gc,
		    TEXT_MARGIN/2, 
		    TEXT_MARGIN + text_height * 2,
		    tbuf);

    gdk_draw_pixmap(disp->window,
		    disp->style->fg_gc[GTK_WIDGET_STATE(disp)],
		    pixmap,
		    0, 0,
		    0, 0,
		    disp->allocation.width,
		    disp->allocation.height);

    return;
  }

  if(orient == ORIENT_LEFT || orient == ORIENT_RIGHT)
    gants_draw_v();
  else
    gants_draw_h();
}

static int gants_get_best_size_v(void)
{
  int text_height, bar_height;
  /*
   * Print job type name on top,
   * then bars for each node.
   */
  load_my_font();

  text_height = gdk_string_height(my_font, active_jobtype ? : "X");
  bar_height = 2 + CELL_SIZE;

  return 2 * TEXT_MARGIN + BAR_MARGIN
    + (usock >= 0 ? 1 : 2) * text_height
    + bar_height * n_metrics;
}

static int gants_get_best_size_h(void)
{
  int text_width, bar_width;
  /*
   * Print job type name on left,
   * then bars for each node.
   */
  load_my_font();

  text_width = gdk_string_width(my_font, active_jobtype ? : "X");
  bar_width = 2 + CELL_SIZE;

  return 2 * TEXT_MARGIN + BAR_MARGIN
    + (usock >= 0 ? 1 : 2) * text_width 
    + bar_width * n_metrics;
}


static void gants_draw_v(void)
{
  /* declarations */
  gchar type_buf[TYPE_LEN];
  int text_height;
  int y_offset;
  metrics_entry_t * mi;

  /* initializations */
  load_my_font();
  strcpy(type_buf, active_jobtype ? : "no job");
  text_height = gdk_string_height(my_font, type_buf);
  mi = metrics;

  /* let's draw ! */
  gdk_gc_set_foreground(gc, &bg_color);
  gdk_draw_rectangle(pixmap, gc, TRUE, 0,0,
		      disp->allocation.width,
		      disp->allocation.height);
  
  gdk_gc_set_foreground(gc, &f_color);
  gdk_draw_string (pixmap, my_font, gc,
		   TEXT_MARGIN/2, 
		   TEXT_MARGIN/2 + text_height,
		   type_buf);

  y_offset = TEXT_MARGIN + text_height;
  
  while(mi) {
    unsigned c;
    unsigned x_offset = BAR_MARGIN;

    /* draw all running jobs */
    for(c = 0; c < mi->running; c++) {
      if(c < mi->preferred)
	gdk_gc_set_foreground(gc, &low_color);
      else 
	gdk_gc_set_foreground(gc, &high_color);         
      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      x_offset += CELL_SIZE;
    }

    /* draw all preallocated jobs */
    for(; c < mi->running + mi->reserved; c++) {
      gdk_gc_set_foreground(gc, &pre_color);
      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      x_offset += CELL_SIZE;
    }

    /* draw all free jobs */
    for(; c < mi->max; c++) {
      gdk_gc_set_foreground(gc, &bg_color);
      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      x_offset += CELL_SIZE;
    }

    gdk_gc_set_foreground(gc, &f_color);
    gdk_draw_rectangle(pixmap, gc, FALSE, BAR_MARGIN, y_offset,
		       x_offset - BAR_MARGIN, CELL_SIZE);

    mi = mi->next;
    y_offset += CELL_SIZE + 2;
  }

  /* get it all out */
  gdk_draw_pixmap(disp->window,
		  disp->style->fg_gc[GTK_WIDGET_STATE(disp)],
		  pixmap,
		  0, 0,
		  0, 0,
		  disp->allocation.width,
		  disp->allocation.height);
}

static void gants_draw_h(void)
{
  /* declarations */
  gchar type_buf[TYPE_LEN];
  int text_height, text_width;
  int x_offset;
  metrics_entry_t * mi;

  /* initializations */
  load_my_font();
  strcpy(type_buf, active_jobtype ? : "no job");
  text_width = gdk_string_width(my_font, type_buf);
  text_height = gdk_string_height(my_font, type_buf);
  mi = metrics;

  /* let's draw ! */
  gdk_gc_set_foreground(gc, &bg_color);
  gdk_draw_rectangle(pixmap, gc, TRUE, 0,0,
		      disp->allocation.width,
		      disp->allocation.height);
  
  gdk_gc_set_foreground(gc, &f_color);
  gdk_draw_string(pixmap, my_font, gc,
		  TEXT_MARGIN/2, 
		  TEXT_MARGIN/2 + text_height,
		  type_buf);

  x_offset = TEXT_MARGIN + text_width;
  
  while(mi) {
    unsigned c;
    unsigned y_offset = disp->allocation.height - BAR_MARGIN;

    /* draw all running jobs */
    for(c = 0; c < mi->running; c++) {
      if(c < mi->preferred)
	gdk_gc_set_foreground(gc, &low_color);
      else 
	gdk_gc_set_foreground(gc, &high_color);         

      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      y_offset -= CELL_SIZE;
    }

    /* draw all preallocated jobs */
    for(; c < mi->running + mi->reserved; c++) {
      gdk_gc_set_foreground(gc, &pre_color);
      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      y_offset -= CELL_SIZE;
    }

    /* draw all free jobs */
    for(; c < mi->max; c++) {
      gdk_gc_set_foreground(gc, &bg_color);
      gdk_draw_rectangle(pixmap, gc, TRUE, x_offset, y_offset, CELL_SIZE, CELL_SIZE);

      y_offset -= CELL_SIZE;
    }

    gdk_gc_set_foreground(gc, &f_color);
    gdk_draw_rectangle(pixmap, gc, FALSE,
		       x_offset, 
		       y_offset,
		       CELL_SIZE,
		       disp->allocation.height - BAR_MARGIN - y_offset);

    mi = mi->next;
    x_offset += CELL_SIZE + 2;
  }

  /* get it all out */
  gdk_draw_pixmap(disp->window,
		  disp->style->fg_gc[GTK_WIDGET_STATE(disp)],
		  pixmap,
		  0, 0,
		  0, 0,
		  disp->allocation.width,
		  disp->allocation.height);

  
}

